{
    "host": [
        {
            "resource": {
                "cpu": 2,
                "disk_gb": 1028,
                "host": "%(host_name)s",
                "memory_mb": 8192,
                "project": "(total)"
            }
        },
        {
            "resource": {
                "cpu": 0,
                "disk_gb": 0,
                "host": "%(host_name)s",
                "memory_mb": 512,
                "project": "(used_now)"
            }
        },
        {
            "resource": {
                "cpu": 0,
                "disk_gb": 0,
                "host": "%(host_name)s",
                "memory_mb": 0,
                "project": "(used_max)"
            }
        }
    ]
}
