{
    "version": {
        "id": "v2.0",
        "links": [
            {
                "href": "%(host)s/v2/",
                "rel": "self"
            },
            {
                "href": "http://docs.openstack.org/",
                "rel": "describedby",
                "type": "text/html"
            }
        ],
        "media-types": [
            {
                "base": "application/json",
                "type": "application/vnd.openstack.compute+json;version=2"
            }
        ],
        "min_version": "",
        "status": "SUPPORTED",
        "updated": "2011-01-21T11:33:21Z",
        "version": ""
    }
}