{
    "volumes": [
        {
            "attachments": [
                {
                    "device": "/",
                    "id": "%(uuid)s",
                    "serverId": "%(uuid)s",
                    "volumeId": "%(uuid)s"
                }
            ],
            "availabilityZone": "zone1:host1",
            "createdAt": "%(strtime)s",
            "displayDescription": "%(volume_desc)s",
            "displayName": "%(volume_name)s",
            "id": "%(uuid)s",
            "metadata": {},
            "size": 100,
            "snapshotId": null,
            "status": "in-use",
            "volumeType": "Backup"
        }
    ]
}
