#!/bin/bash
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

here=`dirname $0`

if test -z "$1"; then
    set -- HEAD
elif test "x$1" = "x-HEAD"; then
    # Simulate behaviour from flake8wrap.sh
    shift
    set -- HEAD~1
fi

git diff --name-only "$@" | $here/run-tests.py
