{
    "volumeAttachments": [
        {
            "device": "%(device)s",
            "serverId": "%(uuid)s",
            "tag": "%(tag)s",
            "volumeId": "%(volume_id)s",
            "delete_on_termination": true,
            "attachment_id": "%(uuid)s",
            "bdm_uuid": "%(uuid)s"
        },
        {
            "device": "%(text)s",
            "serverId": "%(uuid)s",
            "tag": null,
            "volumeId": "%(volume_id2)s",
            "delete_on_termination": false,
            "attachment_id": "%(uuid)s",
            "bdm_uuid": "%(uuid)s"
        }
    ]
}
