{
    "hypervisor": {
        "hypervisor_hostname": "fake-mini",
        "id": %(hypervisor_id)s,
        "state": "up",
        "status": "enabled",
        "uptime": " 08:32:11 up 93 days, 18:25, 12 users,  load average: 0.20, 0.12, 0.14"
    }
}
