{
    "images": [
        {
            "OS-DCF:diskConfig": "AUTO",
            "created": "2011-01-01T01:02:03Z",
            "id": "70a599e0-31e7-49b7-b260-868f441e862b",
            "links": [
                {
                    "href": "%(versioned_compute_endpoint)s/images/70a599e0-31e7-49b7-b260-868f441e862b",
                    "rel": "self"
                },
                {
                    "href": "%(compute_endpoint)s/images/70a599e0-31e7-49b7-b260-868f441e862b",
                    "rel": "bookmark"
                },
                {
                    "href": "http://glance.openstack.example.com/images/70a599e0-31e7-49b7-b260-868f441e862b",
                    "rel": "alternate",
                    "type": "application/vnd.openstack.image"
                }
            ],
            "metadata": {
                "architecture": "x86_64",
                "auto_disk_config": "True",
                "kernel_id": "nokernel",
                "ramdisk_id": "nokernel"
            },
            "minDisk": 0,
            "minRam": 0,
            "name": "fakeimage7",
            "OS-EXT-IMG-SIZE:size": %(int)s,
            "progress": 100,
            "status": "ACTIVE",
            "updated": "2011-01-01T01:02:03Z"
        },
        {
            "created": "2011-01-01T01:02:03Z",
            "id": "155d900f-4e14-4e4c-a73d-069cbf4541e6",
            "links": [
                {
                    "href": "%(versioned_compute_endpoint)s/images/155d900f-4e14-4e4c-a73d-069cbf4541e6",
                    "rel": "self"
                },
                {
                    "href": "%(compute_endpoint)s/images/155d900f-4e14-4e4c-a73d-069cbf4541e6",
                    "rel": "bookmark"
                },
                {
                    "href": "http://glance.openstack.example.com/images/155d900f-4e14-4e4c-a73d-069cbf4541e6",
                    "rel": "alternate",
                    "type": "application/vnd.openstack.image"
                }
            ],
            "metadata": {
                "architecture": "x86_64",
                "kernel_id": "nokernel",
                "ramdisk_id": "nokernel"
            },
            "minDisk": 0,
            "minRam": 0,
            "name": "fakeimage123456",
            "OS-EXT-IMG-SIZE:size": %(int)s,
            "progress": 100,
            "status": "ACTIVE",
            "updated": "2011-01-01T01:02:03Z"
        },
        {
            "created": "2011-01-01T01:02:03Z",
            "id": "a2459075-d96c-40d5-893e-577ff92e721c",
            "links": [
                {
                    "href": "%(versioned_compute_endpoint)s/images/a2459075-d96c-40d5-893e-577ff92e721c",
                    "rel": "self"
                },
                {
                    "href": "%(compute_endpoint)s/images/a2459075-d96c-40d5-893e-577ff92e721c",
                    "rel": "bookmark"
                },
                {
                    "href": "http://glance.openstack.example.com/images/a2459075-d96c-40d5-893e-577ff92e721c",
                    "rel": "alternate",
                    "type": "application/vnd.openstack.image"
                }
            ],
            "metadata": {
                "kernel_id": "nokernel",
                "ramdisk_id": "nokernel"
            },
            "minDisk": 0,
            "minRam": 0,
            "name": "fakeimage123456",
            "OS-EXT-IMG-SIZE:size": %(int)s,
            "progress": 100,
            "status": "ACTIVE",
            "updated": "2011-01-01T01:02:03Z"
        },
        {
            "OS-DCF:diskConfig": "MANUAL",
            "created": "2011-01-01T01:02:03Z",
            "id": "a440c04b-79fa-479c-bed1-0b816eaec379",
            "links": [
                {
                    "href": "%(versioned_compute_endpoint)s/images/a440c04b-79fa-479c-bed1-0b816eaec379",
                    "rel": "self"
                },
                {
                    "href": "%(compute_endpoint)s/images/a440c04b-79fa-479c-bed1-0b816eaec379",
                    "rel": "bookmark"
                },
                {
                    "href": "http://glance.openstack.example.com/images/a440c04b-79fa-479c-bed1-0b816eaec379",
                    "rel": "alternate",
                    "type": "application/vnd.openstack.image"
                }
            ],
            "metadata": {
                "architecture": "x86_64",
                "auto_disk_config": "False",
                "kernel_id": "nokernel",
                "ramdisk_id": "nokernel"
            },
            "minDisk": 0,
            "minRam": 0,
            "name": "fakeimage6",
            "OS-EXT-IMG-SIZE:size": %(int)s,
            "progress": 100,
            "status": "ACTIVE",
            "updated": "2011-01-01T01:02:03Z"
        },
        {
            "created": "2011-01-01T01:02:03Z",
            "id": "c905cedb-7281-47e4-8a62-f26bc5fc4c77",
            "links": [
                {
                    "href": "%(versioned_compute_endpoint)s/images/c905cedb-7281-47e4-8a62-f26bc5fc4c77",
                    "rel": "self"
                },
                {
                    "href": "%(compute_endpoint)s/images/c905cedb-7281-47e4-8a62-f26bc5fc4c77",
                    "rel": "bookmark"
                },
                {
                    "href": "http://glance.openstack.example.com/images/c905cedb-7281-47e4-8a62-f26bc5fc4c77",
                    "rel": "alternate",
                    "type": "application/vnd.openstack.image"
                }
            ],
            "metadata": {
                "kernel_id": "155d900f-4e14-4e4c-a73d-069cbf4541e6",
                "ramdisk_id": null
            },
            "minDisk": 0,
            "minRam": 0,
            "name": "fakeimage123456",
            "OS-EXT-IMG-SIZE:size": %(int)s,
            "progress": 100,
            "status": "ACTIVE",
            "updated": "2011-01-01T01:02:03Z"
        },
        {
            "created": "2011-01-01T01:02:03Z",
            "id": "cedef40a-ed67-4d10-800e-17455edce175",
            "links": [
                {
                    "href": "%(versioned_compute_endpoint)s/images/cedef40a-ed67-4d10-800e-17455edce175",
                    "rel": "self"
                },
                {
                    "href": "%(compute_endpoint)s/images/cedef40a-ed67-4d10-800e-17455edce175",
                    "rel": "bookmark"
                },
                {
                    "href": "http://glance.openstack.example.com/images/cedef40a-ed67-4d10-800e-17455edce175",
                    "rel": "alternate",
                    "type": "application/vnd.openstack.image"
                }
            ],
            "metadata": {
                "kernel_id": "nokernel",
                "ramdisk_id": "nokernel"
            },
            "minDisk": 0,
            "minRam": 0,
            "name": "fakeimage123456",
            "OS-EXT-IMG-SIZE:size": %(int)s,
            "progress": 100,
            "status": "ACTIVE",
            "updated": "2011-01-01T01:02:03Z"
        },
        {
            "created": "2011-01-01T01:02:03Z",
            "id": "76fa36fc-c930-4bf3-8c8a-ea2a2420deb6",
            "links": [
                {
                    "href": "%(versioned_compute_endpoint)s/images/76fa36fc-c930-4bf3-8c8a-ea2a2420deb6",
                    "rel": "self"
                },
                {
                    "href": "%(compute_endpoint)s/images/76fa36fc-c930-4bf3-8c8a-ea2a2420deb6",
                    "rel": "bookmark"
                },
                {
                    "href": "http://glance.openstack.example.com/images/76fa36fc-c930-4bf3-8c8a-ea2a2420deb6",
                    "rel": "alternate",
                    "type": "application/vnd.openstack.image"
                }
            ],
            "metadata": {
                "architecture": "x86_64",
                "kernel_id": "nokernel",
                "ramdisk_id": "nokernel"
            },
            "minDisk": 0,
            "minRam": 0,
            "name": "fakeimage123456",
            "OS-EXT-IMG-SIZE:size": %(int)s,
            "progress": 100,
            "status": "ACTIVE",
            "updated": "2011-01-01T01:02:03Z"
        },
        {
            "created": "2011-01-01T01:02:03Z",
            "id": "%(eph_encryption_id)s",
            "links": [
                {
                    "href": "%(versioned_compute_endpoint)s/images/%(eph_encryption_id)s",
                    "rel": "self"
                },
                {
                    "href": "%(compute_endpoint)s/images/%(eph_encryption_id)s",
                    "rel": "bookmark"
                },
                {
                    "href": "http://glance.openstack.example.com/images/%(eph_encryption_id)s",
                    "rel": "alternate",
                    "type": "application/vnd.openstack.image"
                }
            ],
            "metadata": {
                "hw_ephemeral_encryption": "True"
            },
            "minDisk": 0,
            "minRam": 0,
            "name": "fakeimage123456",
            "OS-EXT-IMG-SIZE:size": %(int)s,
            "progress": 100,
            "status": "ACTIVE",
            "updated": "2011-01-01T01:02:03Z"
        },
        {
            "created": "2011-01-01T01:02:03Z",
            "id": "%(eph_encryption_disabled_id)s",
            "links": [
                {
                    "href": "%(versioned_compute_endpoint)s/images/%(eph_encryption_disabled_id)s",
                    "rel": "self"
                },
                {
                    "href": "%(compute_endpoint)s/images/%(eph_encryption_disabled_id)s",
                    "rel": "bookmark"
                },
                {
                    "href": "http://glance.openstack.example.com/images/%(eph_encryption_disabled_id)s",
                    "rel": "alternate",
                    "type": "application/vnd.openstack.image"
                }
            ],
            "metadata": {
                "hw_ephemeral_encryption": "False"
            },
            "minDisk": 0,
            "minRam": 0,
            "name": "fakeimage123456",
            "OS-EXT-IMG-SIZE:size": %(int)s,
            "progress": 100,
            "status": "ACTIVE",
            "updated": "2011-01-01T01:02:03Z"
        },
        {
            "created": "2011-01-01T01:02:03Z",
            "id": "%(eph_encryption_luks_id)s",
            "links": [
                {
                    "href": "%(versioned_compute_endpoint)s/images/%(eph_encryption_luks_id)s",
                    "rel": "self"
                },
                {
                    "href": "%(compute_endpoint)s/images/%(eph_encryption_luks_id)s",
                    "rel": "bookmark"
                },
                {
                    "href": "http://glance.openstack.example.com/images/%(eph_encryption_luks_id)s",
                    "rel": "alternate",
                    "type": "application/vnd.openstack.image"
                }
            ],
            "metadata": {
                "hw_ephemeral_encryption": "True",
                "hw_ephemeral_encryption_format": "luks"
            },
            "minDisk": 0,
            "minRam": 0,
            "name": "fakeimage123456",
            "OS-EXT-IMG-SIZE:size": %(int)s,
            "progress": 100,
            "status": "ACTIVE",
            "updated": "2011-01-01T01:02:03Z"
        },
        {
            "created": "2011-01-01T01:02:03Z",
            "id": "%(eph_encryption_plain_id)s",
            "links": [
                {
                    "href": "%(versioned_compute_endpoint)s/images/%(eph_encryption_plain_id)s",
                    "rel": "self"
                },
                {
                    "href": "%(compute_endpoint)s/images/%(eph_encryption_plain_id)s",
                    "rel": "bookmark"
                },
                {
                    "href": "http://glance.openstack.example.com/images/%(eph_encryption_plain_id)s",
                    "rel": "alternate",
                    "type": "application/vnd.openstack.image"
                }
            ],
            "metadata": {
                "hw_ephemeral_encryption": "True",
                "hw_ephemeral_encryption_format": "plain"
            },
            "minDisk": 0,
            "minRam": 0,
            "name": "fakeimage123456",
            "OS-EXT-IMG-SIZE:size": %(int)s,
            "progress": 100,
            "status": "ACTIVE",
            "updated": "2011-01-01T01:02:03Z"
        }
    ]
}
