{
    "instance_usage_audit_logs": {
        "hosts_not_run": [
            "samplehost3"
        ],
        "log": {
            "samplehost0": {
                "errors": 1,
                "instances": 1,
                "message": "Instance usage audit ran for host samplehost0, 1 instances in 0.01 seconds.",
                "state": "DONE"
            },
            "samplehost1": {
                "errors": 1,
                "instances": 2,
                "message": "Instance usage audit ran for host samplehost1, 2 instances in 0.01 seconds.",
                "state": "DONE"
            },
            "samplehost2": {
                "errors": 1,
                "instances": 3,
                "message": "Instance usage audit ran for host samplehost2, 3 instances in 0.01 seconds.",
                "state": "DONE"
            }
        },
        "num_hosts": 4,
        "num_hosts_done": 3,
        "num_hosts_not_run": 1,
        "num_hosts_running": 0,
        "overall_status": "3 of 4 hosts done. 3 errors.",
        "period_beginning": "2012-06-01 00:00:00",
        "period_ending": "2012-07-01 00:00:00",
        "total_errors": 3,
        "total_instances": 6
    }
}
