{
    "volume": {
        "status": "in-use",
        "displayDescription": "%(volume_desc)s",
        "availabilityZone": "zone1:host1",
        "displayName": "%(volume_name)s",
        "attachments": [
            { "device": "/",
              "serverId": "%(uuid)s",
              "id": "%(uuid)s",
              "volumeId": "%(uuid)s"
            }
        ],
     "volumeType": "Backup",
     "snapshotId": null,
     "metadata": {},
     "id": "%(uuid)s",
     "createdAt":  "%(strtime)s",
     "size": 100
    }
}
