{
    "volumeAttachments": [
        {
            "device": "%(device)s",
            "id": "%(volume_id)s",
            "serverId": "%(uuid)s",
            "volumeId": "%(volume_id)s"
        },
        {
            "device": "%(text)s",
            "id": "%(volume_id2)s",
            "serverId": "%(uuid)s",
            "volumeId": "%(volume_id2)s"
        }
    ]
}
