{
    "flavors": [
        {
            "id": "1",
            "links": [
                {
                    "href": "%(versioned_compute_endpoint)s/flavors/1",
                    "rel": "self"
                },
                {
                    "href": "%(compute_endpoint)s/flavors/1",
                    "rel": "bookmark"
                }
            ],
            "name": "m1.tiny"
        },
        {
            "id": "2",
            "links": [
                {
                    "href": "%(versioned_compute_endpoint)s/flavors/2",
                    "rel": "self"
                },
                {
                    "href": "%(compute_endpoint)s/flavors/2",
                    "rel": "bookmark"
                }
            ],
            "name": "m1.small"
        },
        {
            "id": "3",
            "links": [
                {
                    "href": "%(versioned_compute_endpoint)s/flavors/3",
                    "rel": "self"
                },
                {
                    "href": "%(compute_endpoint)s/flavors/3",
                    "rel": "bookmark"
                }
            ],
            "name": "m1.medium"
        },
        {
            "id": "4",
            "links": [
                {
                    "href": "%(versioned_compute_endpoint)s/flavors/4",
                    "rel": "self"
                },
                {
                    "href": "%(compute_endpoint)s/flavors/4",
                    "rel": "bookmark"
                }
            ],
            "name": "m1.large"
        },
        {
            "id": "5",
            "links": [
                {
                    "href": "%(versioned_compute_endpoint)s/flavors/5",
                    "rel": "self"
                },
                {
                    "href": "%(compute_endpoint)s/flavors/5",
                    "rel": "bookmark"
                }
            ],
            "name": "m1.xlarge"
        },
        {
            "id": "6",
            "links": [
                {
                    "href": "%(versioned_compute_endpoint)s/flavors/6",
                    "rel": "self"
                },
                {
                    "href": "%(compute_endpoint)s/flavors/6",
                    "rel": "bookmark"
                }
            ],
            "name": "m1.tiny.specs"
        }
    ]
}
