{
    "quota_set": {
        "cores": 20,
        "instances": 20,
        "key_pairs": 100,
        "metadata_items": 128,
        "ram": 51200,
        "server_groups": 10,
        "server_group_members": 10
    }
}
