{
    "events": [
        {
            "code": 200,
            "name": "%(name)s",
            "server_uuid": "%(uuid)s",
            "status": "%(status)s",
            "tag": "%(tag)s"
        }
    ]
}
