{
    "server_groups": [
        {
            "id": "%(id)s",
            "name": "test",
            "policy": "anti-affinity",
            "rules": {"max_server_per_host": 3},
            "members": [],
            "project_id": "6f70656e737461636b20342065766572",
            "user_id": "fake"
        }
   ]
}
