{
    "quota_class_set": {
        "cores": 20,
        "id": "%(set_id)s",
        "injected_file_content_bytes": 10240,
        "injected_file_path_bytes": 255,
        "injected_files": 5,
        "instances": 10,
        "key_pairs": 100,
        "metadata_items": 128,
        "ram": 51200,
        "server_groups": 10,
        "server_group_members": 10
    }
}
