{
    "interfaceAttachments": [
        {
            "fixed_ips": [
                {
                    "ip_address": "%(ip_address)s",
                    "subnet_id": "%(subnet_id)s"
                }
            ],
            "mac_addr": "%(mac_addr)s",
            "net_id": "%(net_id)s",
            "port_id": "%(port_id)s",
            "port_state": "%(port_state)s",
            "tag": "%(tag)s"
        }
    ]
}