{
    "services": [
        {
            "binary": "nova-compute",
            "host": "host1",
            "status": "UNKNOWN"
        },
        {
            "binary": "nova-compute",
            "host": "host2",
            "status": "UNKNOWN"
        }
    ]
}