{
    "quota_class_set": {
        "cores": 20,
        "floating_ips": -1,
        "fixed_ips": -1,
        "id": "%(set_id)s",
        "injected_file_content_bytes": 10240,
        "injected_file_path_bytes": 255,
        "injected_files": 5,
        "instances": 10,
        "key_pairs": 100,
        "metadata_items": 128,
        "ram": 51200,
        "security_group_rules": -1,
        "security_groups": -1
    }
}
