# Copyright (c) 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures


from nova.compute import resource_tracker


class FakeResourceTracker(resource_tracker.ResourceTracker):
    """Version without a DB requirement."""

    def _update(self, context, compute_node, startup=False):
        pass


class RTMockMixin(object):
    def _mock_rt(self, **kwargs):
        if 'spec_set' in kwargs:
            kwargs.update({'autospec': False})
        return self.useFixture(fixtures.MockPatchObject(
            self.compute, 'rt', **kwargs)).mock
