{
    "hypervisors": [
        {
            "id": "%(hypervisor_id)s",
            "hypervisor_hostname": "host2",
            "state": "up",
            "status": "enabled"
        }
    ],
    "hypervisors_links": [
        {
            "href": "http://openstack.example.com/v2.1/6f70656e737461636b20342065766572/os-hypervisors?limit=1&marker=%(hypervisor_id)s",
            "rel": "next"
        }
    ]
}
